######################################################################################################################################################################
#                                                                                                                                                                    #
#                                                                                                                                                                    #
#                                                 THIS CODE CALCULATES THE PERMANENT OF HOMOGENEOUS BLOCK MATRICES                                                   #
#                                                                                                                                                                    #
#                                                       Written by Diego Fasoli (E-mail: diego.fasoli@iit.it)                                                        #
#                                                                                                                                                                    #
#                                          Laboratory of Neural Computation, Center for Neuroscience and Cognitive Systems                                           #
#                                                    @UniTn, Istituto Italiano di Tecnologia, 38068 Rovereto, Italy                                                  #
#                                                                                                                                                                    #
#                                                                      TESTED WITH PYTHON 2.7                                                                        #
#                                                                                                                                                                    #
#                                                                                                                                                                    #
######################################################################################################################################################################



# DESCRIPTION: this code compares the time required by the Balasubramanian-Bax-Franklin-Glynn formula and by Eq. (20) of the main text, for the calculation of the
# permanent of homogeneous block matrices. The code shows that Eq. (20), which is specific for homogeneous block matrices, allows a much faster evaluation of the
# permanent (especially for large matrix sizes N), compared to the Balasubramanian-Bax-Franklin-Glynn formula. This result proves that Eq. (20) allows a much faster
# evaluation of the long-time behavior of statistically-homogeneous multi-population networks, see the main text.
#
# NOTE: here we implemented Eq. (20) only for matrices with one or two block-columns, since these represent the relevant cases for our study (see SubSec. (2.3) of
# the main text).



import numpy as np
from sympy import multinomial_coefficients
import time



######################################################################################################################################################################
######################################################################                          ######################################################################
######################################################################    NETWORK PARAMETERS    ######################################################################
######################################################################                          ######################################################################
######################################################################################################################################################################



N = 22       # Total size of the block-matrix B. Note that B is composed of x_l \times y_m homogeneous blocks B_lm, for l = 0, ..., X-1 and m = 0, 1.
X = 5        # Number of block-rows of the matrix B, i.e. the number of distinct populations that share a given external stimulus.

entries_list = np.around(np.random.uniform(0, 0.3, (X, 2)), decimals=2)      # Collection of the entries of the matrices B_lm, which in this example are two-digits
                                                                             # random numbers generated from a uniform distribution.

y_0 = 3      # Number of columns of the matrices B_l0 (note that 0 <= y_0 <= N, and that the number of columns of the matrices B_l1 is y_1 = N-y_0).
             # When y_0 = 0 or y_0 = N, the matrix B has only one block-column.

x_list = np.random.multinomial(N, np.ones(X)/X, size=1)[0]      # x_list = [x_0, x_1, ..., x_(X-1)], where x_l are random integers that sum to N.



######################################################################################################################################################################
########################################################################                      ########################################################################
########################################################################      FUNCTIONS       ########################################################################
########################################################################                      ########################################################################
######################################################################################################################################################################



def BBFG_permanent(matrix):      #  Balasubramanian-Bax-Franklin-Glynn formula for the calculation of the permanent.
    
    column_sum = matrix.sum(0)
    size = len(matrix)

    permanent = np.prod(column_sum)
    
    v = range(size)
    deltas = np.ones(size)

    sign = 1
    i = 0
    
    while(i < size-1):
        
        column_sum -= matrix[i]*deltas[i]*2
        sign = -sign
        deltas[i] = -deltas[i]
        permanent += sign*np.prod(column_sum)
        
        v[0] = 0
        v[i] = v[i+1]
        v[i+1] = i+1
        i = v[0]
        
    return permanent/(2.**(size-1))



def HBM_permanent(X, x_list, y_0, entries_list, N):      # Permanent of homogeneous block matrices. This is the implementation of Eq. (20) of the main text.

    MC_0 = multinomial_coefficients(X, y_0)
    MC_1 = multinomial_coefficients(X, N-y_0)

    S = MC_0.keys()

    C = np.prod([float(np.math.factorial(x_list[l])) for l in range(X)])      # NOTE: without the float(.) function we may get overflows.
    
    per = C*sum([MC_0[s]*MC_1[tuple(x_list-s)]*np.prod([np.power(entries_list[:,0], s)*np.power(entries_list[:,1], x_list-s)]) for s in S if((x_list >= s).all())])

    return per



######################################################################################################################################################################
##############################################################                                         ###############################################################
##############################################################      CALCULATION OF THE PERMANENT       ###############################################################
##############################################################                                         ###############################################################
######################################################################################################################################################################



column_B_l0 = np.repeat(entries_list[:,0], x_list)
column_B_l1 = np.repeat(entries_list[:,1], x_list)



if(y_0 == 0):

    B = np.array([column_B_l1,]*N)

if(y_0 == N):

    B = np.array([column_B_l0,]*N)

if(0 < y_0 < N):

    B = np.concatenate((np.array([column_B_l0,]*y_0), np.array([column_B_l1,]*(N-y_0))), axis=0)


    
######################################################################################################################################################################
####################################################################                              ####################################################################
####################################################################      PRINT THE RESULTS       ####################################################################
####################################################################                              ####################################################################
######################################################################################################################################################################



print 'Please wait a few seconds...'

time_0 = time.time()

BBFG_per = BBFG_permanent(B)

time_1 = time.time()

print '\nBBFG formula: Permanent = ' + str(BBFG_per) + ' (t = ' + '{:.3f}'.format(time_1 - time_0) + 's)'

HBM_per = HBM_permanent(X, x_list, y_0, entries_list, N)

time_2 = time.time()

print 'Eq. (20): Permanent = ' + str(HBM_per) + ' (t = ' + '{:.3f}'.format(time_2 - time_1) + 's)'

if(0 < time_2-time_1 < time_1-time_0):

    print '\nEq. (20) was ' + '{:.3f}'.format((time_1-time_0)/(time_2-time_1)) + ' times faster than the Balasubramanian-Bax-Franklin-Glynn formula in evaluating the matrix permanent.'

if(0 < time_1-time_0 < time_2-time_1):

    print '\nEq. (20) was ' + '{:.3f}'.format((time_2-time_1)/(time_1-time_0)) + ' times slower than the Balasubramanian-Bax-Franklin-Glynn formula in evaluating the matrix permanent.'

print '\nNOTE: due to rounding errors, a small difference between the output of Eq. (20) and that of the Balasubramanian-Bax-Franklin-Glynn formula is possible, especially when the permanent is large.'

if(HBM_per != 0):
    
    print '\nRelative error between the two methods = ' + '{:.3e}'.format(100*abs(BBFG_per - HBM_per)/HBM_per) + '%'
