
######################################################################################################################################################################
#                                                                                                                                                                    #
#                                                                                                                                                                    #
#                                   THIS CODE COMPUTES THE STATISTICAL PROPERTIES OF THE STATIONARY STATES OF A BINARY RATE MODEL.                                   #
#                                                                                                                                                                    #
#                                                       Written by Diego Fasoli (E-mail: diego.fasoli@iit.it).                                                       #
#                                                                                                                                                                    #
#                                          Laboratory of Neural Computation, Center for Neuroscience and Cognitive Systems                                           #
#                                                   @UniTn, Istituto Italiano di Tecnologia, 38068 Rovereto, Italy.                                                  #
#                                                                                                                                                                    #
#                                                               Part of this code is adapted from:                                                                   #
#                                                                                                                                                                    #
#                    "D. Fasoli and S. Panzeri. Optimized brute-force algorithms for the bifurcation analysis of a binary neural network model"                      #
#                                                                 Phys. Rev. E 99, 012316, 2019                                                                      #
#                                                                                                                                                                    #
#                                                      with permission of the authors. Tested with Python 2.7.                                                       #
#                                                                                                                                                                    #
#                                                                                                                                                                    #
######################################################################################################################################################################



# DESCRIPTION: this code computes semi-analytically the following quantities:

# 1) the cumulative distribution function of the bifurcation points \Lambda and \Xi (see SubSec. (2.2.1) of the main text);
# 2) the mean multistability diagram (SubSec. (2.2.2));
# 3) the occurrence probability of the stationary firing-rate states, for a fixed combination of stimuli (SubSec. (2.2.3));
# 4) the occurrence probability of the stationary firing-rate states, regardless of the stimuli (SubSec. (2.2.4)).
#
# The code also compares these quantities with the corresponding numerical counterparts (see SubSec. (2.5) of the main text).



import numpy as np
from scipy.stats import semicircular
from scipy.ndimage.interpolation import shift
import networkx as nx
import itertools as it
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.colors as colors
from operator import add



######################################################################################################################################################################
######################################################################                          ######################################################################
######################################################################    NETWORK PARAMETERS    ######################################################################
######################################################################                          ######################################################################
######################################################################################################################################################################



N_E = 2            # Size of the excitatory
N_I = 2            # and inhibitory population.


MCR = 5000         # Number of network repetitions in the numerical Monte Carlo simulations.


x_min = -40        # x_min and x_max determine the range of the x_axis, which will be used to plot the cumulative distribution functions of the bifurcation points.
x_max = 40         # NOTE: generally the range [x_min, x_max] should be wide, otherwise the algorithm may miss some mass of the pdf of the total synaptic input
                   # S_{i} = \sum_{j=0}^{N-1}J_{i,j}\nu_{j}, during the calculation of Eq. (7).
delta_x = 0.01     # Resolution of the x_axis. This is also the integration step for the numerical calculation of the 1D definite integrals.


I_E_min = -14      # Boundaries of the mean multistability diagram (i.e. min and max
I_E_max = 15       # currents in the parameters space). NOTE: choose these currents
I_I_min = -11      # so that the parameter space is wide, otherwise discrepancies
I_I_max = 18       # between semi-analytical and numerical results may arise.


theta_E = [0, 1]   # Firing thresholds of the excitatory and inhibitory neurons.
theta_I = [1, 2]   # NOTE: these lists must always contain N_E and N_I entries, respectively.


P = np.array([[0.0, 0.5, 1.0, 0.6],    # Connection probability matrix, ...
              [0.4, 0.5, 0.1, 1.0],
              [0.5, 0.7, 0.3, 0.8],
              [0.0, 1.0, 0.9, 0.0]])


C = np.array([[1, 4, -3, -10],         # ... centers of the Wigner semicircle distribution of the random variables W_ij, ...
              [6, 5, -2,  -4],
              [3, 4, -6,  -7],
              [1, 2, -5,   1]])


R = np.array([[1, 4, 2, 3],            # ... and their corresponding radii, see Tab. (1) of the main text.
              [5, 3, 2, 3],            # NOTE 1: P, C and R must be NxN matrices, where N = N_E + N_I.
              [3, 4, 5, 6],            # NOTE 2: it should be R_ij >> delta_x, otherwise the algorithm
              [1, 2, 4, 1]])           # may miss part of the mass of the distribution of W_ij.
                                       # NOTE 3: in this example we suppose that the variables W_ij are distributed
                                       # according to the Wigner semicircle law. However, any other (continuous)
                                       # distribution can be used as well, if desired.


I_E_fixed = 0   # Fixed stimuli used for the calculation of the occurrence probability
I_I_fixed = 4   # of SubSec. (2.2.3).


chosen_state = [1,1,1,0]   # Chosen firing-rate state among the 2^N possible combinations. The algorithm calculates
                           # and plots the cumulative distribution function of its bifurcation points, according to
                           # Eq. (14) of the main text.
                           # NOTE 1: this state must be a vector of length N = N_E + N_I.
                           # NOTE 2: similarly to the main text, and for the sake of brevity, in this code we plot
                           # only F_Lambda_E and F_Lambda_I. For this reason, note that the bifurcation points of the
                           # state [0,0,0,0] will not be plotted, since this state is stationary only in the set
                           # [-\infty, \Xi_E]x[-\infty, \Xi_I]. However, our code can be extended in order to plot
                           # also F_Xi_E and F_Xi_I, if desired.


'''
# To observe the formation of spontaneous symmetry-breaking in the mean multistability diagram, use for example the following values of the parameters:


theta_E = [1, 1]
theta_I = [2, 2]


P = np.array([[1.0, 0.5, 0.3, 0.3],
              [0.5, 1.0, 0.3, 0.3],
              [0.3, 0.3, 0.0, 0.8],
              [0.3, 0.3, 0.8, 0.0]])


C = np.array([[5, 5, -4, -4],
              [5, 5, -4, -4],
              [5, 5, -4, -4],
              [5, 5, -4, -4]])


R = np.array([[5, 5, 4, 4],
              [5, 5, 4, 4],
              [5, 5, 4, 4],
              [5, 5, 4, 4]])
'''



######################################################################################################################################################################
########################################################################                      ########################################################################
########################################################################     COLORS LIST      ########################################################################
########################################################################                      ########################################################################
######################################################################################################################################################################



# Colors_list is the list of colors to be used for plotting the mean multistability diagram. If colors_list does not contain enough user-defined colors, the algorithm
# will generate random colors for plotting the diagram. NOTE: do not add the color "white" to the list, because it is already reserved for plotting the areas without
# stationary states (namely with multistability degree = 0).


colors_list = ['red', 'green', 'blue', 'cyan', 'yellow', 'magenta', 'orange', 'brown', 'violet', 'pink', '#808080']  # Hexadecimal colors may be used as well.


# The full list of named colors may be found at "http://matplotlib.org/examples/color/named_colors.html".


'''

# The list may be obtained also by running the following two-lines code:

for color_name, color_hex in colors.cnames.iteritems():
    print(color_name, color_hex)

'''



#####################################################################################################################################################################
################################################################                                     ################################################################
################################################################        RANDOM VARIABLES W_ij        ################################################################
################################################################                                     ################################################################
#####################################################################################################################################################################



N = N_E + N_I   # Total size of the network.


n_x = int(float(x_max - x_min)/float(delta_x))
x_axis = np.linspace(x_min, x_max-delta_x, n_x)
W = np.zeros((N, N, n_x), dtype=float)


for i, j in it.product(range(N), repeat=2):

    W[i,j,:] = semicircular.pdf(x_axis, loc = C[i,j], scale = R[i,j])   # Wigner semicircle distribution.


mean_W = np.dot(W, x_axis)*delta_x
mean_W_squared = np.dot(W*W, x_axis)*delta_x



######################################################################################################################################################################
########################################################################                      ########################################################################
########################################################################      FUNCTIONS       ########################################################################
########################################################################                      ########################################################################
######################################################################################################################################################################



def Heaviside(x):   # Heaviside step function. H(x) = 1 for x >= 0, while H(x) = 0 otherwise.
    return 1*(x >= 0)



def permanent(matrix):   # This is a simple implementation of the Balasubramanian-Bax-Franklin-Glynn formula for the calculation of the permanent.

    column_sum = matrix.sum(0)
    permanent = np.prod(column_sum)
    
    size = len(matrix)
    v = range(size)
    deltas = np.ones(size)

    sign = 1
    i = 0
    
    while(i < size-1):
        
        column_sum -= matrix[i]*deltas[i]*2
        sign = -sign
        deltas[i] = -deltas[i]
        permanent += sign*np.prod(column_sum)
        
        v[0] = 0
        v[i] = v[i+1]
        v[i+1] = i+1
        i = v[0]
        
    return permanent/2.**(size-1)



def cdf_and_per(e, population):   # This function returns the cumulative distribution function of the bifurcation points, see Eq. (14), and the permanent in Eq. (13).

    z = np.where(i == 1)[0]

    Gamma_I = np.arange(population*N_E, N_E + population*N_I)
    neural_indexes = np.asarray(Gamma_I)[np.where(i[Gamma_I] == e)[0]]
    L = len(neural_indexes)

    if(L > 0):

        a = np.zeros((L, n_x), dtype=float)
        a_shifted_integral = np.zeros((L, n_x), dtype=float)
        a_shifted = np.zeros((L, n_x), dtype=float)
        b = np.zeros(L, dtype=float)
        cumul_I = np.zeros((L, n_x), dtype=float)

        for m in range(L):

            P_row_reduced = [P[neural_indexes[m],k] for k in z]

            first_term = np.zeros((2**len(z)-1, n_x), dtype=float)

            if(len(z) > 0):

                counter = 0

                for j in it.product((0, 1), repeat=len(z)):

                    j = np.array(j)

                    x = np.where(j == 1)[0]
                    y = np.where(j == 0)[0]
                    
                    if(len(x) > 0):

                        v = np.prod([P_row_reduced[k] for k in x])
                        u = np.prod([1-P_row_reduced[k] for k in y])

                        convolution = W[neural_indexes[m],z[x[0]],:]

                        for n in range(1, len(x)):

                            convolution = np.convolve(convolution, W[neural_indexes[m],z[x[n]],:], 'same')*delta_x   # NOTE: this is equivalent to calculating the convolution by rectangular integration.

                        first_term[counter,:] = u*v*convolution
                        counter += 1

            a[m,:] = np.sum(first_term, axis = 0)
            a_shifted[m,:] = shift(a[m,::-1], int(float(theta_vector[neural_indexes[m]])/float(delta_x)), cval=0)   # NOTE: we set cval = 0 because we suppose that the mass of the pdf is within the range [x_min, x_max], and therefore that the pdf = 0 outside this range.
            a_shifted_integral[m,:] = [np.sum(np.multiply(Heaviside(x_axis - theta_vector[neural_indexes[m]] + s), a[m,:]))*delta_x for s in x_axis]
            b[m] = np.prod([1-Pr for Pr in P_row_reduced])
            cumul_I[m,:] = a_shifted_integral[m,:] + b[m]*Heaviside(x_axis - theta_vector[neural_indexes[m]])

        
        r = 1 + e*(L - 1)

        F_1 = [permanent(np.array([cumul_I[:,t],]*L))/float(np.math.factorial(L)) for t in range(n_x)]         
        F_2 = [sum([permanent(np.concatenate((np.array([cumul_I[:,t],]*q), np.array([1-cumul_I[:,t],]*(L-q))), axis=0))/float(np.math.factorial(q)*np.math.factorial(L-q)) for q in range(r, L)]) for t in range(n_x)]
        F = map(add, F_1, F_2)

        if(L > 1):

            Per = [permanent(np.concatenate((np.array([a_shifted[:,t],]*1), np.array([1-e-((-1)**e)*cumul_I[:,t],]*(L-1))), axis=0))/float(np.math.factorial(L-1)) for t in range(n_x)]

        else:

            Per = a_shifted[0,:]

    else:

        F = 0
        Per = 0

    return F, Per



def mean_bifurcation_points_method_1(e, population, F, Per):   # This function calculates the mean bifurcation points, according to Eq. (15).

    Gamma_I = np.arange(population*N_E, N_E + population*N_I)
    neural_indexes = Gamma_I[np.where(i[Gamma_I] == e)[0]]
    L = len(neural_indexes)

    if(L > 0):

        continuous_part = np.sum(np.multiply(x_axis, Per))*delta_x
        discrete_part = 0

        for q in set(np.array(theta_vector)[neural_indexes]):   # Sum on all the distinct firing thresholds.

            g = int(float(q - x_min)/float(delta_x))
            discrete_part += q*(F[g+1] - F[g-1])

        out = continuous_part + discrete_part

    else:

        out = (e)*((1 - population)*I_E_min + population*I_I_min) + (not e)*((1 - population)*I_E_max + population*I_I_max)

    return out



def mean_bifurcation_points_method_2(e, population, F):         # This is an implementation of Eq. (16), which represents an alternative method for calculating the mean bifurcation points.
                                                                # NOTE: method_2 generally provides more precise results than method_1, since the calculation of F[g+1] - F[g-1] in method_1
    Gamma_I = np.arange(population*N_E, N_E + population*N_I)   # is very sensitive to numerical errors, and tipically requires smaller values of delta_x in order to provide accurate results.
    neural_indexes = Gamma_I[np.where(i[Gamma_I] == e)[0]]
    L = len(neural_indexes)

    if(L > 0):

        out = np.sum(Heaviside(x_axis) - F)*delta_x

    else:

        out = (e)*((1 - population)*I_E_min + population*I_I_min) + (not e)*((1 - population)*I_E_max + population*I_I_max)

    return out



def P_for_fixed_stimuli(population, F_0, F_1):   # This function returns P\left(\widehat{I}_{\alpha}\in\mathcal{V}_{\alpha}\right), see Eq. (17).

    Gamma_I = np.arange(population*N_E, N_E + population*N_I)
    L = len(Gamma_I[np.where(i[Gamma_I] == 0)[0]])

    n = (1 - population)*N_E + population*N_I

    if(L == n):

        prob = 1-F_0

    if(L == 0):

        prob = F_1

    if(0 < L < n):

        prob = (1-F_0)*F_1

    return prob



def P_regardless_of_the_stimuli(population, F_0, F_1, Per):   # This function returns P\left(\mathrm{len}\left(\mathcal{V}_{\alpha}\right)>0\right), see Eq. (18).

    Gamma_I = np.arange(population*N_E, N_E + population*N_I)

    neural_indexes_0 = Gamma_I[np.where(i[Gamma_I] == 0)[0]]
    L_0 = len(neural_indexes_0)

    neural_indexes_1 = Gamma_I[np.where(i[Gamma_I] == 1)[0]]
    L_1 = len(neural_indexes_1)

    if(L_0 > 0 and L_1 > 0):
        
        probability_continuous_part = np.sum(np.multiply(F_0, Per))*delta_x
        probability_discrete_part = 0

        for q in set(np.array(theta_vector)[neural_indexes_0]):

            g = int(float(q - x_min)/float(delta_x))
            probability_discrete_part += F_0[g+1]*(F_1[g+1] - F_1[g-1])

        return probability_continuous_part + probability_discrete_part

    else:

        return 1



def symmetry_breaking(i, x_rect_min, x_rect_max , y_rect_min, y_rect_max, rect_list, list_E, list_I):   # This function determines the areas of the mean multistability diagram where
                                                                                                        # spontaneous symmetry-breaking eventually occurred.
    if(x_rect_min < x_rect_max and y_rect_min < y_rect_max):   # (x_rect_min, x_rect_max, y_rect_min, y_rect_max) are the coordinates of the vertices of the rectangles.
                                                               # If x_rect_min >= x_rect_max and/or y_rect_min >= y_rect_max, the corresponding stationary state is not allowed.
        rect_list.append((x_rect_min, y_rect_min, x_rect_max, y_rect_max))
                
        sym_break_E = 0
        sym_break_I = 0

        if(equations_are_symmetric and N_E > 1 and 0 < np.sum(i[0:N_E]) < N_E):

            sym_break_E = 1

        if(equations_are_symmetric and N_I > 1 and 0 < np.sum(i[N_E:N]) < N_I):

            sym_break_I = 1

        list_E.append(sym_break_E)
        list_I.append(sym_break_I)



def create_array(population, y, z):   # This function creates the arguments of the min and max functions, according to Eq. (4).

    x = np.arange(population*N_E, N_E + population*N_I)
    indexes = x[np.where(y[x] == z)[0]]

    if(len(indexes) > 0):

        my_array = theta_vector[indexes] - np.dot(J[indexes,:], y)

    else:

        my_array = []

    return my_array



def plot_cdf(F_semianalytical, F_numerical, y_axis_label, figure_name, x_axis_min, x_axis_max):   # Plot the cumulative distribution functions of the bifurcation points.

    plt.figure(figure_name)

    plt.plot(x_axis, F_semianalytical, 'r-', linewidth=3)
    plt.plot(x_axis, F_numerical, 'bo', markersize = 10, markevery = 120)

    plt.xlabel(r'$x$', fontsize=35)
    plt.ylabel(y_axis_label, fontsize=30)

    plt.axhline(y = 0, linewidth = 3, linestyle = '--', color = 'k', label='_nolegend_')
    plt.axhline(y = 1, linewidth = 3, linestyle = '--', color = 'k', label='_nolegend_')

    plt.legend( ('Semi-analytical', 'Numerical'), loc='best', prop={'size':18} )
    plt.xlim([x_axis_min, x_axis_max])
    plt.ylim([-0.2, 1.2])



def rgb_color(my_color):   # This function converts the named/hexadecimal colors contained in colors_list into their corresponding RGB colors.

    col = my_color

    if(my_color[0] != '#'):
        col = colors.cnames[my_color]

    return colors.hex2color(col)



def intersection(rect_0, rect_1):   # This function determines whether two rectangles overlap in the mean multistability diagram, by checking if the intersection area is positive.
       
    intersection_area = max(0, min(rect_0[2], rect_1[2]) - max(rect_0[0], rect_1[0]))*max(0, min(rect_0[3], rect_1[3]) - max(rect_0[1], rect_1[1]))

    return intersection_area



def plot(ax, my_dictionary, my_edgecolor, my_hatch):   # Plot the rectangles that form the mean multistability diagram.

    legend_sym_break_E = 0
    legend_sym_break_I = 0
    legend_sym_break_EI = 0

    for i in my_dictionary.keys():

        j = my_dictionary[i][0]
        
        while(j > len(colors_list) - 1):   # If the user has not defined enough colors in colors_list, then add some random colors.
            while(True):                   # The code makes an attempt to find distinct random colors.

                random_color = tuple(np.random.random_integers(0, 255, 3)/255.)

                if(np.prod([(random_color != colors_list[k]) for k in range(0, len(colors_list))]) and random_color != (0, 0, 0)):

                    colors_list.append(random_color)
                    break
        
        my_hatch_ = None

        if(my_hatch != None):

            if(my_dictionary[i][1] == True and my_dictionary[i][2] == False):
                my_hatch_ = '\\'
                legend_sym_break_E = 1

            if(my_dictionary[i][1] == False and my_dictionary[i][2] == True):
                my_hatch_ = '/'
                legend_sym_break_I = 1

            if(my_dictionary[i][1] == True and my_dictionary[i][2] == True):
                my_hatch_ = 'x'
                legend_sym_break_EI = 1

        ax.add_patch(patches.Rectangle(
                (i[0], i[1]),
                 i[2] - i[0],
                 i[3] - i[1],
                 alpha = 1.0,
                 facecolor = colors_list[j],
                 edgecolor = my_edgecolor,
                 hatch = my_hatch_
        )
        )

    return legend_sym_break_E, legend_sym_break_I, legend_sym_break_EI



def plot_overlapping_areas(ax, my_rect_dict, my_cliques_list, my_edgecolor, my_hatch):   # Plot the overlapping areas of the rectangles with a new color, depending on their degrees of multistability.

    for i in my_cliques_list:

        new_dict= {}
        rect_0 = my_rect_dict.keys()[i[0]]
        counter = my_rect_dict[rect_0][0]
        sym_E = my_rect_dict[rect_0][1]
        sym_I = my_rect_dict[rect_0][2]

        for j in range(1, len(i)):

            rect_1 = my_rect_dict.keys()[i[j]]
            counter += my_rect_dict[rect_1][0]
            sym_E += my_rect_dict[rect_1][1]
            sym_I += my_rect_dict[rect_1][2]
            rect_0 = (max(rect_0[0], rect_1[0]), max(rect_0[1], rect_1[1]), min(rect_0[2], rect_1[2]), min(rect_0[3], rect_1[3]))

        new_dict[rect_0] = (counter, (sym_E > 0), (sym_I > 0))
        plot(ax, new_dict, my_edgecolor, my_hatch)   # Plot the rectangles sequentially, in order to assure that the overlapping areas are plotted on the lower-degree rectangles.



def plot_mean_multistability_diagram(rect_list, list_E, list_I, figure_name):   # Plot the mean multistability diagram on the (I_E, I_I) plane.


    ##### STEP 1: plot the rectangles of the mean multistability diagram without edges, since they give rise to spurious colors. This problem is caused by the color interpolation performed by matplotlib when drawing the canvas.


    S = set(rect_list)
    L = [[(sum([list_E[rect_list.index(j)] for j in rect_list if j == i]) > 0), (sum([list_I[rect_list.index(j)] for j in rect_list if j == i]) > 0)] for i in S]
    rect_dict = dict((tup, (rect_list.count(tup), L[list(S).index(tup)][0], L[list(S).index(tup)][1])) for tup in S)   # rect_list.count(tup) represents the degree of multistability of the area delimited by the rectangle, which determines its color.

    fig = plt.figure(num=figure_name, facecolor='black')
    ax = fig.add_subplot(111, aspect='equal')
    ax.axis([I_E_min, I_E_max, I_I_min, I_I_max])
    ax.axis('off')
    

    plot(ax, {(I_E_min, I_I_min, I_E_max, I_I_max): (0, 0, 0)}, 'none', None)
    plot(ax, rect_dict, 'none', None)   # Plot the rectangles in rect_list, which will appear in the background of the mean multistability diagram.


    n_rect = len(rect_dict)
    A = np.zeros((n_rect, n_rect), dtype=float)
    w = rect_dict.keys()


    for i in w:           # If the mth and nth rectangles overlap, then A[m,n] = 1, otherwise A[m,n] = 0.
        for j in w:       # This allows us to reinterpret the rectangles and their intersections as the nodes and edges of a directed graph
            if(i != j):   # respectively, which proves very convenient for assigning the color to each rectangle.

                A[w.index(i),w.index(j)] = (intersection(i, j) > 0)


    G = nx.from_numpy_matrix(A)


    cliques_list = [s for s in nx.enumerate_all_cliques(G) if len(s) > 1]   # Find all the fully-connected subgraphs (i.e. the cliques) of G. For example, the presence
                                                                            # of a 3x3 clique means that there are three rectangles that overlap with
                                                                            # a common intersection area.
    plot_overlapping_areas(ax, rect_dict, cliques_list, 'none', None)


    ##### STEP 2: find all the visible colors in the diagram.


    fig.canvas.draw()
    rgb_data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
    rgb_data = rgb_data.reshape((int(len(rgb_data)/3), 3))/255.
    rgb_data = [tuple(i) for i in np.vstack({tuple(row) for row in rgb_data})]
    rgb_data.remove((0, 0, 0))


    degrees_list = sorted([colors_list.index(a) for a in rgb_data])


    ##### STEP 3: "erase" the figure and plot it again by including the rectangle's edges, the axes, the legend, and the areas where symmetry-breaking eventually occurred.


    plt.cla()


    font = {'size'   : 22}
    plt.rc('font', **font)
    fig.set_facecolor('silver')
    legend_sym_break_E, legend_sym_break_I, legend_sym_break_EI = plot(ax, rect_dict, 'black', 'Plot With Hatch')
    plot_overlapping_areas(ax, rect_dict, cliques_list, 'black', 'Plot With Hatch')
    plt.xlabel(r'$I_{E}$', fontsize=35)
    plt.ylabel(r'$I_{I}$', fontsize=35)
    plt.xlim(xmax = I_E_max, xmin = I_E_min)
    plt.ylim(ymax = I_I_max, ymin = I_I_min)


    legend_list_0 = []

    for i in list(degrees_list):

        legend_list_0.append(patches.Patch(facecolor = colors_list[i], label = str(i), edgecolor = 'black'))

    legend_0 = plt.legend(handles = legend_list_0, loc = 'center left', bbox_to_anchor = (1, 0.5), prop = {'size':15}, ncol = 2)


    legend_list_1 = []

    if(legend_sym_break_E):
        legend_list_1.append(patches.Patch(facecolor = 'white', label = 'Broken symmetry in the excitatory population', edgecolor = 'black', hatch = '\\\\'))

    if(legend_sym_break_I):
        legend_list_1.append(patches.Patch(facecolor = 'white', label = 'Broken symmetry in the inhibitory population', edgecolor = 'black', hatch = '//'))

    if(legend_sym_break_EI):
        legend_list_1.append(patches.Patch(facecolor = 'white', label = 'Broken symmetry in both the populations', edgecolor = 'black', hatch = 'xx'))
        
    if(legend_sym_break_E or legend_sym_break_I or legend_sym_break_EI):
        plt.legend(handles = legend_list_1, loc = 'upper center', bbox_to_anchor = (0.5, 1.12), prop = {'size':12})
        plt.gca().add_artist(legend_0)



def plot_occurrence_probability(P_semianalytical, P_numerical, y_axis_label, figure_name):   # Plot the occurrence probability of the stationary states, for fixed stimuli or regardless of them.

    plt.figure(figure_name)

    x_labels_list = [str('{0:b}'.format(i)).zfill(N) for i in range(2**N)]

    ax = plt.subplot(111)
    width = 1
    bins = map(lambda x: x-width, range(1, 2**N+1))
    ax.bar(bins, P_semianalytical, width=width/2., color='r')
    ax.bar(np.array(bins)+width/2., P_numerical, width=width/2., color='b')
    ax.set_xticks(map(lambda x: x-width/2., range(1, 2**N+1)))
    ax.set_xticklabels(x_labels_list, rotation=90, rotation_mode='anchor', ha='right')

    plt.xlim([0, 2**N])
    plt.ylim([0, 1])
    plt.ylabel(y_axis_label, size = 30)
    plt.legend( ('Semi-analytical', 'Numerical'), loc='best', prop={'size':15} )



def binary_to_decimal(x):   # Return the decimal representation of the binary number x.

    return sum(int(x[i])*(2**(N-i-1)) for i in range(N))
        



######################################################################################################################################################################
##############################################################                                          ##############################################################
##############################################################            CHECK FOR SYMMETRY            ##############################################################
##############################################################                                          ##############################################################
######################################################################################################################################################################



# The code checks whether the neural populations are statistically homogeneous (see SubSec. (2.2.2) of the main text).


equations_are_symmetric = 0


mean_sum = [np.sum(P[i[0]*N_E:N_E+i[0]*N_I,i[1]*N_E:N_E+i[1]*N_I]*mean_W[i[0]*N_E:N_E+i[0]*N_I,i[1]*N_E:N_E+i[1]*N_I], axis = 1) for i in it.product((0, 1), repeat=2)]
var_sum = [np.sum(P[i[0]*N_E:N_E+i[0]*N_I,i[1]*N_E:N_E+i[1]*N_I]*mean_W_squared[i[0]*N_E:N_E+i[0]*N_I,i[1]*N_E:N_E+i[1]*N_I] - pow(P[i[0]*N_E:N_E+i[0]*N_I,i[1]*N_E:N_E+i[1]*N_I]*mean_W[i[0]*N_E:N_E+i[0]*N_I,i[1]*N_E:N_E+i[1]*N_I], 2), axis = 1) for i in it.product((0, 1), repeat=2)]
check_mean = [np.size(np.unique(np.around(i, decimals=6))) for i in mean_sum]
check_var = [np.size(np.unique(np.around(i, decimals=6))) for i in var_sum]


check_theta_E = np.size(np.unique(np.around(theta_E, decimals=6)))
check_theta_I = np.size(np.unique(np.around(theta_I, decimals=6)))


if(sum(check_mean + check_var) == 8 and check_theta_E + check_theta_I == 2):

    equations_are_symmetric = 1



######################################################################################################################################################################
###########################################################                                                ###########################################################
###########################################################          SEMI-ANALYTICAL CALCULATIONS          ###########################################################
###########################################################                                                ###########################################################
######################################################################################################################################################################



theta_vector = np.array(theta_E + theta_I)


rect_list_semianalytical = []


symmetry_breaking_E_list_semianalytical = []
symmetry_breaking_I_list_semianalytical = []


F = np.zeros((4, n_x), dtype=float)
Per = np.zeros((4, n_x), dtype=float)


P_fixed_stimuli_semianalytical = np.zeros(2**N, dtype=float)
P_regardless_stimuli_semianalytical = np.zeros(2**N, dtype=float)


counter = 0


print 'Please wait, semi-analytical calculation in progress ...\n'


for i in it.product((0, 1), repeat=N):   # Check the 2^N firing-rate states.  


    i = np.array(i)


    ##### Mean multistability diagram.


    F[0,:], Per[0,:] = cdf_and_per(1, 0)
    x_rect_min = mean_bifurcation_points_method_1(1, 0, F[0,:], Per[0,:])   # Calculate the mean bifurcation points by means of Eq. (15) (method_1) ...
    # x_rect_min = mean_bifurcation_points_method_2(1, 0, F[0,:])           # ... or Eq. (16) (method_2).

    F[1,:], Per[1,:] = cdf_and_per(0, 0)
    x_rect_max = mean_bifurcation_points_method_1(0, 0, F[1,:], Per[1,:])
    # x_rect_max = mean_bifurcation_points_method_2(0, 0, F[1,:])

    F[2,:], Per[2,:] = cdf_and_per(1, 1)
    y_rect_min = mean_bifurcation_points_method_1(1, 1, F[2,:], Per[2,:])
    # y_rect_min = mean_bifurcation_points_method_2(1, 1, F[2,:])

    F[3,:], Per[3,:] = cdf_and_per(0, 1)
    y_rect_max = mean_bifurcation_points_method_1(0, 1, F[3,:], Per[3,:])
    # y_rect_max = mean_bifurcation_points_method_2(0, 1, F[3,:])


    ##### Cumulative distribution function of the bifurcation points for the chosen firing-rate state.


    if(np.array_equal(i, chosen_state)):

        F_Lambda_E_semianalytical = np.copy(F[0,:])
        F_Lambda_I_semianalytical = np.copy(F[2,:])
        F_Xi_E_semianalytical = np.copy(F[1,:])
        F_Xi_I_semianalytical = np.copy(F[3,:])


    ##### Occurrence probability of the stationary states for fixed stimuli.


    input_E = int(float(I_E_fixed - x_min)/float(delta_x))
    input_I = int(float(I_I_fixed - x_min)/float(delta_x))

    P_fixed_stimuli_semianalytical[counter] = P_for_fixed_stimuli(0, F[1,input_E], F[0,input_E])*P_for_fixed_stimuli(1, F[3,input_I], F[2,input_I])


    ##### Occurrence probability of the stationary states regardless of the stimuli.


    P_regardless_stimuli_semianalytical[counter] = P_regardless_of_the_stimuli(0, F[0,:], F[1,:], Per[1,:])*P_regardless_of_the_stimuli(1, F[2,:], F[3,:], Per[3,:])
    

    ##### Find the areas of the mean multistability diagram where spontaneous symmetry-breaking eventually occurred.


    symmetry_breaking(i, x_rect_min, x_rect_max, y_rect_min, y_rect_max, rect_list_semianalytical, symmetry_breaking_E_list_semianalytical, symmetry_breaking_I_list_semianalytical)


    #####


    counter += 1


print 'Semi-analytical calculation completed.'



#######################################################################################################################################################################
##############################################################                                           ##############################################################
##############################################################           NUMERICAL SIMULATIONS           ##############################################################
##############################################################                                           ##############################################################
#######################################################################################################################################################################



print '\n\n\nPlease wait, numerical calculation in progress ...\n'


rect_list_numerical = []


symmetry_breaking_E_list_numerical = []
symmetry_breaking_I_list_numerical = []


x_rect_min_average = np.zeros(2**N, dtype=float)
x_rect_max_average = np.zeros(2**N, dtype=float)
y_rect_min_average = np.zeros(2**N, dtype=float)
y_rect_max_average = np.zeros(2**N, dtype=float)


p_Lambda_E_numerical = np.zeros(n_x, dtype=float)
p_Lambda_I_numerical = np.zeros(n_x, dtype=float)


P_fixed_stimuli_numerical = np.zeros(2**N, dtype=float)
P_regardless_stimuli_numerical = np.zeros(2**N, dtype=float)


for s in range(MCR):   # Monte Carlo simulation.
    

    counter = 0

    J = np.multiply(semicircular.rvs(loc=C, scale=R, size=(N, N)), np.random.binomial(1, P))   # Generate a quenched connectivity matrix at every network realization.


    for i in it.product((0, 1), repeat=N):


        i = np.array(i)


        ##### Mean multistability diagram.


        x_rect_min = -10**10
        x_rect_max =  10**10
        y_rect_min = -10**10
        y_rect_max =  10**10
        
        a = create_array(0, i, 1)
        b = create_array(0, i, 0)
        c = create_array(1, i, 1)
        d = create_array(1, i, 0)

        if(np.size(a) > 0):

            x_rect_min = np.amax(a)

        if(np.size(b) > 0):

            x_rect_max = np.amin(b)
            
        if(np.size(c) > 0):

            y_rect_min = np.amax(c)

        if(np.size(d) > 0):

            y_rect_max = np.amin(d)

        x_rect_min_average[counter] += x_rect_min
        y_rect_min_average[counter] += y_rect_min
        x_rect_max_average[counter] += x_rect_max
        y_rect_max_average[counter] += y_rect_max


        ##### Probability density of the bifurcation points for the chosen firing-rate state. This density will be used for the numerical calculation of the corresponding cumulative distribution function.


        if(np.array_equal(i, chosen_state)):

            address = [int((x_rect_min-x_min)/delta_x), int((y_rect_min-x_min)/delta_x)]

            if(0 <= address[0] < n_x):

                p_Lambda_E_numerical[address[0]] += 1

            if(0 <= address[1] < n_x):

                p_Lambda_I_numerical[address[1]] += 1

            # Similarly for the probability distributions p_Xi_E_numerical and p_Xi_I_numerical.


        ##### Occurrence probability of the stationary states for fixed stimuli.


        if(x_rect_min <= I_E_fixed < x_rect_max and y_rect_min <= I_I_fixed < y_rect_max):   # Alternatively, this line of code could be replaced by "if(np.array_equal(i, Heaviside(np.dot(J, i) + np.concatenate((I_E_fixed*np.ones(N_E, dtype=float), I_I_fixed*np.ones(N_I, dtype=float))) - theta_vector))):",
                                                                                             # which checks the stationarity condition \nu_i(t+1) = \nu_i(t) of the firing-rate states directly on Eq. (1), for a sinchronously updated network (see SubSec. (2.5)). Note that synchronous and asynchronous
            P_fixed_stimuli_numerical[counter] += 1                                          # networks have the same set of stationary solutions, therefore this "if" statement can be used also in the case when the neurons are updated asynchronously.


        ##### Occurrence probability of the stationary states regardless of the stimuli.


        if(x_rect_min <= x_rect_max and y_rect_min <= y_rect_max):

            P_regardless_stimuli_numerical[counter] += 1


        #####
            

        counter += 1


F_Lambda_E_numerical = np.cumsum(p_Lambda_E_numerical)/float(MCR)
F_Lambda_I_numerical = np.cumsum(p_Lambda_I_numerical)/float(MCR)


x_rect_min_average = x_rect_min_average/float(MCR)
y_rect_min_average = y_rect_min_average/float(MCR)
x_rect_max_average = x_rect_max_average/float(MCR)
y_rect_max_average = y_rect_max_average/float(MCR)


P_fixed_stimuli_numerical = P_fixed_stimuli_numerical/float(MCR)
P_regardless_stimuli_numerical = P_regardless_stimuli_numerical/float(MCR)


counter = 0


for i in it.product((0, 1), repeat=N):


    i = np.array(i)
                                

    ##### Find the areas of the mean multistability diagram where spontaneous symmetry-breaking eventually occurred.


    symmetry_breaking(i, x_rect_min_average[counter], x_rect_max_average[counter], y_rect_min_average[counter], y_rect_max_average[counter], rect_list_numerical, symmetry_breaking_E_list_numerical, symmetry_breaking_I_list_numerical)


    counter += 1


print 'Numerical calculation completed.'



######################################################################################################################################################################
###########################################################                                                ###########################################################
###########################################################                PLOT THE RESULTS                ###########################################################
###########################################################                                                ###########################################################
######################################################################################################################################################################



print '\n\n\nPlotting the results ...'


font = {'size'   : 18}
plt.rc('font', **font)


##### Cumulative distribution function of the bifurcation points.


Gamma_I = np.arange(N_E)


if(len(np.where(np.array(chosen_state)[Gamma_I] == 1)[0]) > 0):

    plot_cdf(F_Lambda_E_semianalytical, F_Lambda_E_numerical, r'$F_{\Lambda_{E}}\left(x\right)$', 'Cumulative distribution function of Lambda_E', -15, 15)


Gamma_I = np.arange(N_E, N)


if(len(np.where(np.array(chosen_state)[Gamma_I] == 1)[0]) > 0):

    plot_cdf(F_Lambda_I_semianalytical, F_Lambda_I_numerical, r'$F_{\Lambda_{I}}\left(x\right)$', 'Cumulative distribution function of Lambda_I', -15, 15)


# Similarly for F_Xi_E and F_Xi_I.


##### Mean multistability diagram.


colors_list = [rgb_color(colors_list[h]) for h in range(0, len(colors_list))]
colors_list.insert(0, (1, 1, 1))   # (1, 1, 1) = "white". This is the color that represents the areas without stationary states (namely with multistability degree = 0).

plot_mean_multistability_diagram(rect_list_semianalytical, symmetry_breaking_E_list_semianalytical, symmetry_breaking_I_list_semianalytical, 'Mean multistability diagram (semi-analytical)')
plot_mean_multistability_diagram(rect_list_numerical, symmetry_breaking_E_list_numerical, symmetry_breaking_I_list_numerical, 'Mean multistability diagram (numerical)')


##### Occurrence probability of the stationary states, for fixed stimuli or regardless of them.


plot_occurrence_probability(P_fixed_stimuli_semianalytical, P_fixed_stimuli_numerical, r'$P\left(\widehat{I}\in\mathscr{V}\right)$', 'Occurrence probability for fixed stimuli')
plot_occurrence_probability(P_regardless_stimuli_semianalytical, P_regardless_stimuli_numerical, r'$P\left(\mathrm{vol}\left(\mathscr{V}\right)>0\right)$', 'Occurrence probability regardless of the stimuli')


plt.show()
